# -*- coding: utf-8 -*-


import numpy as np
from typing import Dict


def load_entity_to_descriptions(path_to_file: str) -> np.array:
    """Load entity to descriptions file."""

    entity_description_matrix = np.loadtxt(
        fname=path_to_file,
        dtype=str,
        comments='@Comment@ Subject Predicate Object',
        delimiter='\t',
    )

    return entity_description_matrix


def create_entity_to_desciption_mappings(entity_description_matrix: np.array) -> Dict[str, str]:
    """Map entities to descriptions"""

    entities = entity_description_matrix[:, 0]
    descriptions = entity_description_matrix[:, 1]

    entity_to_desciption = {}

    for i in range(len(entities)):
        entity = entities[i]
        desc = descriptions[i]

        # Check whether entity/description is empty or None
        if not entity or entity is None or not desc or desc is None:
            continue

        entity_to_desciption[entity] = desc

    return entity_to_desciption



# -*- coding: utf-8 -*-

import numpy as np
from typing import Dict

from src.kupp.triples_preprocessing_utils.basic_triple_utils import slice_triples


def create_multi_label_relation_instances(unique_entity_pairs: np.array,
                                          triples: np.array,
                                          num_relations: int,
                                          create_class_other=False
                                          ) -> Dict[tuple, np.array]:
    """Create for each (s,o) pair the multi relation label."""

    subjects, relations, objects = slice_triples(triples)
    entity_pairs = np.concatenate([subjects, objects], axis=1)

    # Create class 'other' for relations not contained in the KG
    if create_class_other:
        num_relations += 1

    s_t_to_multi_relations = create_multi_label_instances(unique_pairs=unique_entity_pairs,
                                                          pairs=entity_pairs,
                                                          elements=relations,
                                                          num_elements=num_relations)
    return s_t_to_multi_relations


def create_multi_label_objects_instance(unique_s_r_pairs: np.array,
                                        triples: np.array,
                                        num_entities: int, ) -> Dict[tuple, np.array]:
    """Create for each (s,r) pair the multi object label."""

    subjects, relations, objects = slice_triples(triples)
    s_r_pairs = np.concatenate([subjects, relations], axis=1)

    s_r_to_mulit_objects = create_multi_label_instances(unique_pairs=unique_s_r_pairs,
                                                        pairs=s_r_pairs,
                                                        elements=objects,
                                                        num_elements=num_entities)
    return s_r_to_mulit_objects


def create_multi_label_instances(unique_pairs: np.array,
                                 pairs: np.array,
                                 elements: np.array,
                                 num_elements) -> Dict[tuple, np.array]:
    """Create for each (element_1, element_2) pair the multi-label."""

    instance_to_multi_label = {}

    for unique_pair in unique_pairs:
        # Step 1: Get all corresponding elements of pair
        indices = np.where((pairs == unique_pair).all(-1))
        all_corresponding_elements_of_pair = np.array(np.sort(elements[indices]).tolist(), dtype=np.int)
        # Step 2: Create hot encoding labels
        multi_label = np.zeros(num_elements)
        np.put(multi_label, all_corresponding_elements_of_pair, np.ones(len(all_corresponding_elements_of_pair)))
        # Step 3: Save in dict
        instance_to_multi_label[tuple(unique_pair)] = multi_label

    return instance_to_multi_label



# -*- coding: utf-8 -*-

"""Abstract class for pre-trained models that encode text passages."""

import numpy as np
from abc import ABC, abstractmethod

class AbstractSentenceEncoder(ABC):

    @abstractmethod
    def encode(self, texts:list) -> np.array:
        """Encode the text passages into vector representations."""


# -*- coding: utf-8 -*-

"""Utils to process and embed text passages."""

import multiprocessing
from multiprocessing import Pool

import numpy as np
import spacy
from typing import Dict

nlp = spacy.load('en')


def _extract_sentences(descriptions: np.array) -> list:
    """Extract first n sentence of description."""
    processed_descriptions = []

    for description in descriptions:
        # Remove quotations marks from the beginning and the end
        if description.startswith('"') and description.endswith('"'):
            description = description[1:-1]

        sentences = []
        # doc is a container
        doc = nlp(str(description))
        for i, sent in enumerate(doc.sents):
            sentences.append(sent.text)
            if i == MAX_NUM_SENTENCES - 1:
                break

        sentences = " ".join(sentences)
        processed_descriptions.append(sentences)

    return processed_descriptions


def extract_first_n_sentences(entity_to_desc: Dict[str, str],
                              max_num_sentences: int,
                              num_processes=multiprocessing.cpu_count()) -> Dict[str, str]:
    """Extract first n sentences of description, and return mapping of entity to processed descriptions."""
    global MAX_NUM_SENTENCES
    MAX_NUM_SENTENCES = max_num_sentences

    entities = list(entity_to_desc.keys())
    descriptions = list(entity_to_desc.values())
    description_chunks = np.array_split(descriptions, num_processes)

    with Pool(num_processes) as p:
        # Order of results is same as order description_chunks
        # results is a list of lists where each sublist represent the result of a process
        results = p.map(_extract_sentences, description_chunks)

    processed_decs = []

    for r in results:
        processed_decs += r

    entity_to_desc = dict(zip(entities, processed_decs))

    return entity_to_desc


def embedd_texts(texts:list) -> np.array:
    """"""
    pass


# -*- coding: utf-8 -*-

"""Test utils for processing numeric literals."""

import unittest

class NumericLiteralsUtilsTests(unittest.TestCase):
    """Class for testing utils for processing numeric literals."""

    def test_create_matix_of_literals(self):
        """."""
        self.assertEqual(1,1)


# -*- coding: utf-8 -*-


from typing import Dict, Optional, Tuple

import numpy as np


def load_triples(path, delimiter='\t'):
    """Load triples saved as tab separated values."""
    triples = np.loadtxt(
        fname=path,
        dtype=str,
        comments='@Comment@ Subject Predicate Object',
        delimiter=delimiter,
    )
    return triples


def create_entity_and_relation_mappings(triples: np.array) -> Tuple[Dict[str, int], Dict[str, int]]:
    """Map entities and relations to ids."""
    entities = np.unique(np.ndarray.flatten(np.concatenate([triples[:, 0:1], triples[:, 2:3]])))
    relations = np.unique(np.ndarray.flatten(triples[:, 1:2]).tolist())

    entity_to_id: Dict[int, str] = {
        value: key
        for key, value in enumerate(entities)
    }

    rel_to_id: Dict[int, str] = {
        value: key
        for key, value in enumerate(relations)
    }

    return entity_to_id, rel_to_id


def create_triple_mappings(triples: np.array, are_triples_unique=True) -> Dict[tuple, int]:
    """Create mappings for triples."""

    if not are_triples_unique:
        triples = np.unique(ar=triples, axis=0)

    triples_to_id: Dict[tuple, int] = {
        tuple(value): key
        for key, value in enumerate(triples)
    }

    return triples_to_id


def map_triples_elements_to_ids(triples: np.array,
                                entity_to_id: Optional[Dict[int, str]] = None,
                                rel_to_id: Optional[Dict[int, str]] = None) -> np.ndarray:
    """Map entities and relations to predefined ids."""

    subject_column = np.vectorize(entity_to_id.get)(triples[:, 0:1])
    relation_column = np.vectorize(rel_to_id.get)(triples[:, 1:2])
    object_column = np.vectorize(entity_to_id.get)(triples[:, 2:3])
    triples_of_ids = np.concatenate([subject_column, relation_column, object_column], axis=1)

    triples_of_ids = np.array(triples_of_ids, dtype=np.long)
    # Note: Unique changes the order of the triples
    return np.unique(ar=triples_of_ids, axis=0), entity_to_id, rel_to_id


def get_unique_entity_pairs(triples, return_indices=False) -> np.array:
    """Extract all unique entity pairs from the triples."""

    subjects, _, objects = slice_triples(triples)

    entity_pairs = np.concatenate([subjects, objects], axis=1)

    return get_unique_pairs(pairs=entity_pairs, return_indices=return_indices)


def get_unique_subject_relation_pairs(triples, return_indices=False) -> np.array:
    """Extract all unique subject relation pairs from the triples."""

    subjects, relations, _ = slice_triples(triples)

    subject_relation_pairs = np.concatenate([subjects, relations], axis=1)

    return get_unique_pairs(pairs=subject_relation_pairs, return_indices=return_indices)


def get_unique_pairs(pairs, return_indices=False) -> np.array:
    """Extract unique pairs."""

    # idx: Indices in triples of unique pairs
    _, idx = np.unique(pairs, return_index=True, axis=0)
    sorted_indices = np.sort(idx)
    # uniquoe pairs where original order of triples is preserved
    unique_pairs = pairs[sorted_indices]

    if return_indices:
        return unique_pairs, sorted_indices
    return unique_pairs


def slice_triples(triples):
    """Get the heads, relations, and tails from a matrix of triples."""

    heads = triples[:, 0:1]
    relations = triples[:, 1:2]
    triples = triples[:, 2:3]

    return heads, relations, triples



# -*- coding: utf-8 -*-

"""Wrapper for the universal sentence encoder."""
from src.kupp.text_preprocessing_utils.AbstractSentenceEncoder import AbstractSentenceEncoder
import tensorflow_hub as hub
import tensorflow as tf
import numpy as np


class USEncoder(AbstractSentenceEncoder):
    def __init__(self):
        # model 2 corresponds to the deep averaging network
        self.pretrained_model = hub.Module("https://tfhub.dev/google/universal-sentence-encoder/2")

    def encode(self, texts: list) -> np.array:
        """Encode text passages."""

        embeddings = self.pretrained_model(texts)

        with tf.Session() as sess:
            sess.run(tf.global_variables_initializer())
            sess.run(tf.tables_initializer())
            embedded_sentences = np.array(sess.run(embeddings))

        return embedded_sentences



# -*- coding: utf-8 -*-

"""Utils to process numerical literals."""

import numpy as np
from typing import Dict
import logging
log = logging.getLogger(__name__)

def create_matix_of_literals(triples: np.array, entity_to_id: Dict) -> np.array:
    """Create matrix of literals where each row corresponds to an entity and each column to a literal."""
    data_rel_to_id : Dict[str, int] = {
        value: key
        for key, value in enumerate(triples[:,2:3])
    }

    # Prepare literal matrix, set every literal to zero, and afterwards fill in the corresponding value if available
    num_literals = np.zeros([len(entity_to_id), len(data_rel_to_id)], dtype=np.float32)

    # ToDo: Vecotrize code
    for i, (h, r, lit) in enumerate(triples):
        try:
            # row define entity, and column the literal. Set the corresponding literal for the entity
            num_literals[entity_to_id[h], data_rel_to_id[r]] = lit
        except KeyError:
            log.info("Either entity or relation to literal doesn't exist.")
            continue

    return num_literals